import numpy as np
import opfunu  
import random
import math


def get_cec_fun(func_num, dim):
    func = opfunu.get_functions_by_classname(func_num)[0](ndim=dim)
    return func


class cs:
    def __init__(self, func_num, dim, epoch=6, pop_size=4):
        self.item = epoch  
        self.pop_size = pop_size
        self.func = get_cec_fun(func_num, dim)
        self.p_num = dim  
        self.lb = list(self.func.lb)
        self.ub = list(self.func.ub)

    def search(self, folder_path=None):
        pa = 0.25
        nest = np.zeros((self.pop_size, self.p_num))
        for i in range(self.p_num):
            nest[:, i] = np.random.uniform(0, 1, self.pop_size) * (self.ub[i] - self.lb[i]) + self.lb[i]
        new_nest = np.copy(nest)
        fitness = np.zeros(self.pop_size)
        fitness.fill(float("inf"))
        fmin, bestnest, nest, fitness = self.get_best_nest(nest, new_nest, fitness)
        convergence = []
        for iter in range(0, self.item):
            new_nest = self.get_cuckoos(nest, bestnest)
            fnew, best, nest, fitness = self.get_best_nest(nest, new_nest, fitness)
            new_nest = self.empty_nests(new_nest, pa)
            for j in range(0, self.pop_size):
                s2 = new_nest[j, :]
                for k in range(self.p_num):
                    new_nest[j, k] = np.clip(s2[k], self.lb[k], self.ub[k])
            fnew, best, nest, fitness = self.get_best_nest(nest, new_nest, fitness)
            if fnew < fmin:
                fmin = fnew
                bestnest = best
            convergence.append(fmin)
        convergences = np.asarray(convergence)
        min_epoch = np.where(convergences == fmin)[0][0] + 1
        if folder_path is not None:
            f = open(folder_path + 'fitness.txt', 'a')
            f.write('**************************** start search ***********************************')
            f.write('fitness:{}'.format(convergence))
            f.write('\n')
            f.write('\n')
            f.close()
        return bestnest, fmin, min_epoch

    def cec_fun_best(self):
        return self.func.x_global, self.func.f_global, self.func.n_fe

    def fun(self, x):
        t = self.func.evaluate(x)
        return t

    def get_cuckoos(self, nest, best):
        tempnest = np.array(nest)
        beta = 3 / 2
        sigma = (
            math.gamma(1 + beta)
            * math.sin(math.pi * beta / 2)
            / (math.gamma((1 + beta) / 2) * beta * 2 ** ((beta - 1) / 2))
        ) ** (1 / beta)
        history = []
        for j in range(0, self.pop_size):
            history.append(nest[j, :])
            s = nest[j, :]
            u = np.random.randn(len(s)) * sigma
            v = np.random.randn(len(s))
            step = u / abs(v) ** (1 / beta)
            stepsize = 0.01 * (step * (s - best))
            s = s + stepsize * np.random.randn(len(s))
            for k in range(self.p_num):
                if s[k] < self.lb[k] or s[k] > self.ub[k]:
                    s[k] = np.random.uniform(self.lb[k], self.ub[k])
                    tempnest[j, k] = s[k]
                    history[j] = tempnest[j, :]
        return tempnest



    def get_best_nest(self, nest, newnest, fitness):
        tempnest = np.copy(nest)
        for j in range(0, self.pop_size):
            fnew = self.fun(newnest[j, :])
            if fnew <= fitness[j]:
                fitness[j] = fnew
                tempnest[j, :] = newnest[j, :]
        fmin = min(fitness)
        K = np.argmin(fitness)
        bestlocal = tempnest[K, :]
        return fmin, bestlocal, tempnest, fitness


    def empty_nests(self, nest, pa):
        K = np.random.uniform(0, 1, (self.pop_size, self.p_num)) > pa
        stepsize = random.random() * (nest[np.random.permutation(self.pop_size), :] - nest[np.random.permutation(self.pop_size), :])
        tempnest = nest + stepsize * K
        return tempnest