import numpy as np
import opfunu  
import random


def get_cec_fun(func_num, dim):
    func = opfunu.get_functions_by_classname(func_num)[0](ndim=dim)
    return func


class ics_zoa:
    def __init__(self, func_num, dim, epoch=6, pop_size=4):
        self.item = epoch  
        self.pop_size = pop_size
        self.func = get_cec_fun(func_num, dim)
        self.p_num = dim  
        self.lb = list(self.func.lb)
        self.ub = list(self.func.ub)

    def search(self, folder_path=None):
        pa_min = 0.15
        pa_max = 0.55
        alpha = 1.0
        beta = 1.0
        lam = 0.1
        nest = self.initialize_nest()
        new_nest = np.copy(nest)
        fitness = np.full(self.pop_size, float("inf"))
        convergence = []
        fmin, bestnest, tempnest, fitness = self.get_best_nest(nest, new_nest, fitness)
        for iter in range(0, self.item):
            for j in range(0, self.pop_size):
                I = int(round(1 + np.random.random()))
                tempnest[j, :] = np.random.random() * np.abs((bestnest - I * tempnest[j, :]))
                for k in range(self.p_num):
                    if tempnest[j, k] < self.lb[k] or tempnest[j, k] > self.ub[k]:
                        tempnest[j, k] = np.random.uniform(self.lb[k], self.ub[k])
            history = []
            new_nest = tempnest
            fmin, bestnest, tempnest, fitness = self.get_best_nest(nest, new_nest, fitness)
            for j in range(0, self.pop_size):
                history.append(tempnest[j, :])
                s = tempnest[j, :]
                d = j
                while d == j:
                    d = np.random.randint(self.pop_size)
                Az = np.copy(tempnest[d, :])
                Ps = np.random.random()
                if Ps < 0.5:
                    R = 0.1
                    s = s + R * (2 * np.random.random() - 1) * (1 - iter / self.item)
                else:
                    I = int(round(1 + np.random.random()))  
                    s = s + np.random.random() * (Az - I * s)
                for k in range(self.p_num):
                    if s[k] < self.lb[k] or s[k] > self.ub[k]:
                        s[k] = np.random.uniform(self.lb[k], self.ub[k])
                        tempnest[j, k] = s[k]
                        history[j] = tempnest[j, :]
            new_nest = tempnest
            fnew, best, nest, fitness = self.get_best_nest(nest, new_nest, fitness)
            pa = self.discovery_probability(pa_min, pa_max, alpha, beta, lam)
            new_nest = self.empty_nests(nest, pa)
            new_nest = np.clip(new_nest, self.lb, self.ub)
            fnew, best, nest, fitness = self.get_best_nest(nest, new_nest, fitness)
            if fnew < fmin:
                fmin = fnew
                bestnest = best
            convergence.append(fmin)
        convergences = np.asarray(convergence)
        min_epoch = np.where(convergences == fmin)[0][0]+1
        if folder_path is not None:
            f = open(folder_path+'fitness.txt', 'a')
            f.write('**************************** start search ***********************************')
            f.write('fitness:{}'.format(convergence))
            f.write('\n')
            f.write('\n')
            f.close()
        return bestnest, fmin, min_epoch

    def cec_fun_best(self):
        return self.func.x_global, self.func.f_global, self.func.n_fe

    def fun(self, x):
        t = self.func.evaluate(x)
        return t

    @staticmethod
    def tent_map(x, a=1.8):
        if x < 0.5:
            return a * x
        else:
            return a * (1 - x)

    def initialize_nest(self):
        nest = np.zeros((self.pop_size, self.p_num))
        for i in range(self.pop_size):
            for j in range(self.p_num):
                x = np.random.random()
                nest[i, j] = self.lb[j] + (self.ub[j] - self.lb[j]) * self.tent_map(x)
        return nest


    def get_best_nest(self, nest, new_nest, fitness):
        tempnest = np.copy(nest)
        for j in range(0, self.pop_size):
            fnew = self.fun(new_nest[j, :])
            if fnew <= fitness[j]:
                fitness[j] = fnew
                tempnest[j, :] = new_nest[j, :]
        fmin = min(fitness)
        K = np.argmin(fitness)
        bestnest = tempnest[K, :]
        return fmin, bestnest, tempnest, fitness


    def empty_nests(self, nest, pa):
        K = np.random.uniform(0, 1, (self.pop_size, self.p_num)) > pa
        stepsize = random.random() * (nest[np.random.permutation(self.pop_size), :] - nest[np.random.permutation(self.pop_size), :])
        tempnest = nest + stepsize * K
        return tempnest

    @staticmethod
    def discovery_probability(pa_min, pa_max, alpha, beta, lam):
        pa = pa_min + (pa_max - pa_min) * random.random() - lam * np.random.gamma(alpha, beta)
        return pa