# -*- coding: utf-8 -*-
"""
CEC 测试函数 系列
"""

from optimizer.ics_zoa import ics_zoa
from optimizer.cs import cs
from optimizer.zoa import zoa
import numpy as np
import os
import pandas as pd


'''
适应度函数及维度dim的选择
cec函数名字格式：函数名+年份，比如要选择2022的F1函数，func_num = 'F1'+'2022'
cec2019：F1-F10, 可选 dim: F1=9,F2=16,F3=18,其他=10
'''
optim = {
        0: 'ics_zoa',
        1: 'cs',
        2: 'zoa',
    }


def cec_test(optim_num=0, itr=30, year='2019', ipfun=None, ps=50, epoch=100, save_table=False):
    # 单次测试
    global cec_table
    if itr == 1 and ipfun is not None:
        fun_name = 'F' + str(ipfun)  # 按需修改
        func_num = fun_name + year
        dim = select_dim(ipfun)
        model = eval(optim[optim_num])(func_num, dim, epoch=epoch, pop_size=ps)
        bp, bf, me = model.search()
        mp, bv, sn = model.cec_fun_best()
        print('min fitness:{}, the number of search:{}, min position:{}'.format(bv, sn, mp))
        print('best fitness:{}，min epoch:{}, best position:{}'.format(bf, me, bp))
    # 单cec19函数测试
    elif ipfun is not None:
        fun_name = 'F' + str(ipfun)  # 按需修改
        func_num = fun_name + year
        # 维度，根据cec函数 选择对应维度
        dim = select_dim(ipfun)
        run_cec(func_num, ps, epoch, itr, dim, optim_num)
    else:
        if save_table:
            cec_table = pd.DataFrame(data=None, columns=['best', 'worst', 'mean', 'std', 'min_epoch', 'search_num'])
        # cec19函数批量测试
        for i in range(1, 11):
            if i == 12:
                continue
            fun_name = 'F' + str(i)  # 按需修改
            func_num = fun_name + year
            # 维度，根据cec函数 选择对应维度
            dim = select_dim(i)
            if save_table:
                cec_table.loc[i] = run_cec(func_num, ps, epoch, itr, dim, optim_num)
            else:
                run_cec(func_num, ps, epoch, itr, dim, optim_num)
        if save_table:
            cec_table.to_csv('./results/'+optim[optim_num]+'/'+str(ps)+'_'+str(epoch)+'.csv')
    return


def select_dim(a):
    if a == 1:
        dim = 9
    elif a == 2:
        dim = 16
    elif a == 3:
        dim = 18
    else:
        dim = 10
    return dim


def run_cec(func_num, ps, epoch, itr, dim, optim_num):
    temp_list = []
    # result save
    setting = func_num + '_' + str(ps) + '_' + str(epoch)
    folder_path = './results/' + optim[optim_num] + '/' + setting + '/'
    for j in range(itr):
        model = eval(optim[optim_num])(func_num, dim, epoch=epoch, pop_size=ps)
        if not os.path.exists(folder_path + str(j + 1) + '_search/'):
            os.makedirs(folder_path + str(j + 1) + '_search/')
        bp, bf, me = model.search(folder_path + str(j + 1) + '_search/')
        # print('[{}]: Search finish!  {}'.format(j+1, bf))
        mp, bv, sn = model.cec_fun_best()
        temp_list.append([bf, me, sn])
    temp_list = np.array(temp_list)
    best = np.min(temp_list[:, 0])
    worst = np.max(temp_list[:, 0])
    mean = np.mean(temp_list[:, 0])
    std = np.std(temp_list[:, 0])
    min_epoch = np.mean(temp_list[:, 1])
    sn = np.mean(temp_list[:, 2])
    f = open(folder_path + 'optim_result.txt', 'a')
    f.write(func_num + '-' + "{}\n".format(optim[optim_num]))
    f.write('min_position:{}'.format(mp))
    f.write('\n')
    f.write('best_values:{}, search_num:{}'.format(bv, sn))
    f.write('\n')
    f.write('best:{}, worst:{}, mean:{}, std:{}, min_epoch:{}'.format(best, worst, mean, std, min_epoch))
    # f.write('best_fitness, fitness_num, best_position:\n')
    # f.write('{}'.format(temp_list))
    f.write('\n')
    f.write('\n')
    f.close()
    return [best, worst, mean, std, min_epoch, sn]


if __name__ == '__main__':
    import argparse
    # import os
    # os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"

    parser = argparse.ArgumentParser(description='cec19 test')
    parser.add_argument('--sa', type=int, default=0, help='search algorithm')
    parser.add_argument('--fun', type=int, default=None, help='function')
    parser.add_argument('--ps', type=int, default=50, help='pop_size')
    parser.add_argument('--epoch', type=int, default=1000, help='epoch')
    parser.add_argument('--itr', type=int, default=30, help='itr')
    parser.add_argument('--save_table', type=bool, default=True, help='save table')
    args = parser.parse_args()
    cec_test(optim_num=args.sa, itr=args.itr, ipfun=args.fun, ps=args.ps, epoch=args.epoch, save_table=args.save_table)